#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <linux/rtc.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>
#include <sched.h>
#include <pthread.h>

typedef signed long long int cycle_time;
typedef double us_time;
typedef double sec_time;

extern int tracing;
extern int num_threads;
extern double MHZ;
extern int CLOCK_HZ;
extern struct thrd_data_t *thrd_data;
extern struct trace_rec_t *trace_recs;
extern int current_record;
extern cycle_time start, end;
extern cycle_time GAP;
extern int *gap_hist;

enum thrd_task_t { LAT, TRACE };
enum thrd_pri_t { IDLE, LOW, NORMAL, HIGH, HIGHEST, MAX };
enum thrd_status_t { INIT, READY, GO, STOP, DONE };

#define MAXSTRLEN 512
#define LAT_SLOTS 50000
#define NUM_TRACE_RECORDS 100000
#define NUM_GAP_HIST_BUCKETS 1000
#define GAP_HIST_DIVISOR 10

struct thrd_data_t {
  // generic stuff
  volatile enum thrd_status_t status;
  enum thrd_task_t task; 
  enum thrd_pri_t pri;
  pthread_t thrd;
  int id;

  // latency test stuff
  cycle_time *lat_slots;
  int lat_slot;

  // execution trace stuff
  cycle_time last_exec;
  cycle_time exec_start;
  cycle_time thread_time;
  int reserved_record;
};

struct trace_rec_t {
  int id;
  cycle_time start, end;
};

/*
 * rdtsc: this works with gcc on Pentium and Athlon class machines
 *
 * a bit of documentation since this function seems to often be
 * misunderstood:
 *
 * rdtsc is an instruction and "=A" (d) tells gcc that we're
 * outputting a 64-bit quantity, stored in eax:edx, and putting
 * it into variable d
 *
 * when optimization is turned on, gcc does the right thing: it
 * inlines this function and (properly) eliminates storage location d 
 */
static inline cycle_time rdtsc (void)
{
  cycle_time d;
  __asm__ __volatile__ ("rdtsc" : "=A" (d) );
  return d;
}

static inline sec_time cycle_to_sec (cycle_time c)
{
  return c / (sec_time)(MHZ * 1000000);
}

static inline cycle_time sec_to_cycle (sec_time s)
{
  return (cycle_time)s * (cycle_time)(MHZ * 1000000);
}

static inline cycle_time us_to_cycle (us_time us)
{
  return (cycle_time)us * (cycle_time)MHZ;
}

static inline us_time cycle_to_us (cycle_time c)
{
  return c / (us_time)MHZ;
}

extern void *xmalloc (size_t size);

extern void set_priority (thrd_pri_t pri);
extern void lock_pages (void);
extern void setup_fast_clock (void);
extern int RTCLOCKFD;
extern void unset_fast_clock (void);
extern double find_mhz (void);

extern void *thrd_func (void *arg);

extern void print_all_results (void);
