#include "hourglass.h"

void set_priority (thrd_pri_t pri)
{
  int max_pri, res;
  struct sched_param param;

  max_pri = sched_get_priority_max (SCHED_RR);
  if (max_pri == -1) {
    perror ("sched_get_priority_max");
    exit (errno);
  }

  switch (pri) {
  case NORMAL:
    param.sched_priority = 0;
    
    res = sched_setscheduler (0, SCHED_OTHER, &param);
    if (res != 0) {
      perror ("sched_setscheduler");
      exit (errno);
    }
    break;
    
  case HIGHEST:
  case MAX:
    if (pri == HIGHEST) {
      param.sched_priority = max_pri-1;
    } else {
      param.sched_priority = max_pri;
    }
    
    res = sched_setscheduler (0, SCHED_RR, &param);
    if (res != 0) {
      perror ("sched_setscheduler");
      exit (errno);
    }
    break;
    
  default:
    printf ("Hrm: unexpected arg to set_priority: %d\n", pri);
    exit (-1);
  }
}

void lock_pages (void)
{
  int res;

  res = mlockall (MCL_CURRENT | MCL_FUTURE);
  if (res != 0) {
    perror ("mlockall");
    exit (errno);
  }
}


int RTCLOCKFD;

void setup_fast_clock (void)
{
  int retval;

  RTCLOCKFD = open ("/dev/rtc", O_RDONLY);
  
  if (RTCLOCKFD ==  -1) {
    perror ("/dev/rtc");
    exit (errno);
  }
  
  /* Set periodic interrupt frequency */
  retval = ioctl (RTCLOCKFD, RTC_IRQP_SET, CLOCK_HZ);
  if (retval == -1) {
    perror ("ioctl");
    exit (errno);
  }

  /* Enable periodic interrupts */
  retval = ioctl (RTCLOCKFD, RTC_PIE_ON, 0);
  if (retval == -1) {
    perror("ioctl");
    exit(errno);
  }
}

void unset_fast_clock (void)
{
  close (RTCLOCKFD);
}

double find_mhz (void)
{
  /* various strategies are possible here; for Linux we just look in
     /proc, for Win32 we look at rate of change of system time
     vs. rdtsc; could also borrow LCM code from LMBench */
 
  int res;
  double freq;
  char str[MAXSTRLEN], *s2;
  FILE *f = fopen ("/proc/cpuinfo", "r");

  if (!f) {
    printf ("configuration error: could not open /proc/cpuinfo\n");
    printf ("in order to determine processor MHZ\n");
    exit (-1);
  }

  // uncool - this is fragile with respect to the format string...

  do {
    s2 = fgets (str, MAXSTRLEN, f);
    res = sscanf (str, "cpu MHz	: %lf", &freq);
  } while (res == 0 && s2);

  if (res == 0) {
    printf ("Hrm: failed to find MHz in /proc/cpuinfo\n");
    printf ("there was probably a configuration error\n");
    exit (-1);
  }

  fprintf (stderr, "according to /proc/cpuinfo, this is a %f MHz processor\n",
	   freq);
  
  return freq;
}

